@@ -23,16 +23,19 @@ from .api_pb2 import (  # type: ignore
     BluetoothConnectionsFreeResponse,
     BluetoothDeviceConnectionResponse,
     BluetoothDeviceRequest,
+    BluetoothGATTErrorResponse,
     BluetoothGATTGetServicesDoneResponse,
     BluetoothGATTGetServicesRequest,
     BluetoothGATTGetServicesResponse,
     BluetoothGATTNotifyDataResponse,
     BluetoothGATTNotifyRequest,
+    BluetoothGATTNotifyResponse,
     BluetoothGATTReadDescriptorRequest,
     BluetoothGATTReadRequest,
     BluetoothGATTReadResponse,
     BluetoothGATTWriteDescriptorRequest,
     BluetoothGATTWriteRequest,
+    BluetoothGATTWriteResponse,
     BluetoothLEAdvertisementResponse,
     ButtonCommandRequest,
     CameraImageRequest,
@@ -93,7 +96,12 @@ from .api_pb2 import (  # type: ignore
     TextSensorStateResponse,
 )
 from .connection import APIConnection, ConnectionParams
-from .core import APIConnectionError, TimeoutAPIError
+from .core import (
+    APIConnectionError,
+    BluetoothGATTAPIError,
+    TimeoutAPIError,
+    to_human_readable_address,
+)
 from .host_resolver import ZeroconfInstanceType
 from .model import (
     APIVersion,
@@ -102,6 +110,7 @@ from .model import (
     BluetoothConnectionsFree,
     BluetoothDeviceConnection,
     BluetoothDeviceRequestType,
+    BluetoothGATTError,
     BluetoothGATTRead,
     BluetoothGATTServices,
     BluetoothLEAdvertisement,
@@ -408,6 +417,33 @@ class APIClient:
             SubscribeHomeassistantServicesRequest(), on_msg
         )
 
+    async def _send_bluetooth_message_await_response(
+        self,
+        address: int,
+        handle: int,
+        request: message.Message,
+        response_type: Type[message.Message],
+        timeout: float = 10.0,
+    ) -> message.Message:
+        self._check_authenticated()
+        assert self._connection is not None
+
+        def is_response(msg: message.Message) -> bool:
+            return (
+                isinstance(msg, (BluetoothGATTErrorResponse, response_type))
+                and msg.address == address  # type: ignore[union-attr]
+                and msg.handle == handle  # type: ignore[union-attr]
+            )
+
+        resp = await self._connection.send_message_await_response_complex(
+            request, is_response, is_response, timeout=timeout
+        )
+
+        if isinstance(resp[0], BluetoothGATTErrorResponse):
+            raise BluetoothGATTAPIError(BluetoothGATTError.from_pb(resp[0]))
+
+        return resp[0]
+
     async def subscribe_bluetooth_le_advertisements(
         self, on_bluetooth_le_advertisement: Callable[[BluetoothLEAdvertisement], None]
     ) -> Callable[[], None]:
@@ -483,11 +519,20 @@ class APIClient:
             async with async_timeout.timeout(timeout):
                 await event.wait()
         except asyncio.TimeoutError as err:
-            unsub()
-            raise TimeoutAPIError("Timeout waiting for connect response") from err
-        finally:
-            if not event.is_set():
-                await self.bluetooth_device_disconnect(address)
+            try:
+                unsub()
+            except ValueError:
+                _LOGGER.warning(
+                    "%s: Bluetooth device connection timed out but already unsubscribed",
+                    to_human_readable_address(address),
+                )
+            # Disconnect before raising the exception to ensure
+            # the slot is recovered before the timeout is raised
+            # to avoid race were we run out even though we have a slot.
+            await self.bluetooth_device_disconnect(address)
+            raise TimeoutAPIError(
+                f"Timeout waiting for connect response while connecting to {to_human_readable_address(address)} after {timeout}s"
+            ) from err
 
         return unsub
 
@@ -508,10 +553,21 @@ class APIClient:
         self._check_authenticated()
 
         def do_append(msg: message.Message) -> bool:
-            return isinstance(msg, BluetoothGATTGetServicesResponse)
+            return (
+                isinstance(
+                    msg, (BluetoothGATTGetServicesResponse, BluetoothGATTErrorResponse)
+                )
+                and msg.address == address
+            )
 
         def do_stop(msg: message.Message) -> bool:
-            return isinstance(msg, BluetoothGATTGetServicesDoneResponse)
+            return (
+                isinstance(
+                    msg,
+                    (BluetoothGATTGetServicesDoneResponse, BluetoothGATTErrorResponse),
+                )
+                and msg.address == address
+            )
 
         assert self._connection is not None
         resp = await self._connection.send_message_await_response_complex(
@@ -522,58 +578,56 @@ class APIClient:
         )
         services = []
         for msg in resp:
+            if isinstance(msg, BluetoothGATTErrorResponse):
+                raise BluetoothGATTAPIError(BluetoothGATTError.from_pb(msg))
+
             services.extend(BluetoothGATTServices.from_pb(msg).services)
+
         return ESPHomeBluetoothGATTServices(address=address, services=services)
 
     async def bluetooth_gatt_read(
         self,
         address: int,
-        characteristic_handle: int,
+        handle: int,
         timeout: float = DEFAULT_BLE_TIMEOUT,
     ) -> bytearray:
-        self._check_authenticated()
-
         req = BluetoothGATTReadRequest()
         req.address = address
-        req.handle = characteristic_handle
+        req.handle = handle
 
-        def is_response(msg: message.Message) -> bool:
-            if isinstance(msg, BluetoothGATTReadResponse):
-                read = BluetoothGATTRead.from_pb(msg)
-
-                if read.address == address and read.handle == characteristic_handle:
-                    return True
-            return False
-
-        assert self._connection is not None
-        resp = await self._connection.send_message_await_response_complex(
-            req, is_response, is_response, timeout=timeout
+        resp = await self._send_bluetooth_message_await_response(
+            address,
+            handle,
+            req,
+            BluetoothGATTReadResponse,
+            timeout=timeout,
         )
 
-        if len(resp) != 1:
-            raise APIConnectionError(f"Expected one result, got {len(resp)}")
-
-        read_response = BluetoothGATTRead.from_pb(resp[0])
+        read_response = BluetoothGATTRead.from_pb(resp)
 
         return bytearray(read_response.data)
 
     async def bluetooth_gatt_write(
         self,
         address: int,
-        characteristic_handle: int,
+        handle: int,
         data: bytes,
         response: bool,
+        timeout: float = DEFAULT_BLE_TIMEOUT,
     ) -> None:
-        self._check_authenticated()
-
         req = BluetoothGATTWriteRequest()
         req.address = address
-        req.handle = characteristic_handle
+        req.handle = handle
         req.response = response
         req.data = data
 
-        assert self._connection is not None
-        await self._connection.send_message(req)
+        await self._send_bluetooth_message_await_response(
+            address,
+            handle,
+            req,
+            BluetoothGATTWriteResponse,
+            timeout=timeout,
+        )
 
     async def bluetooth_gatt_read_descriptor(
         self,
@@ -581,29 +635,19 @@ class APIClient:
         handle: int,
         timeout: float = DEFAULT_BLE_TIMEOUT,
     ) -> bytearray:
-        self._check_authenticated()
-
         req = BluetoothGATTReadDescriptorRequest()
         req.address = address
         req.handle = handle
 
-        def is_response(msg: message.Message) -> bool:
-            if isinstance(msg, BluetoothGATTReadResponse):
-                read = BluetoothGATTRead.from_pb(msg)
-
-                if read.address == address and read.handle == handle:
-                    return True
-            return False
-
-        assert self._connection is not None
-        resp = await self._connection.send_message_await_response_complex(
-            req, is_response, is_response, timeout=timeout
+        resp = await self._send_bluetooth_message_await_response(
+            address,
+            handle,
+            req,
+            BluetoothGATTReadResponse,
+            timeout=timeout,
         )
 
-        if len(resp) != 1:
-            raise APIConnectionError(f"Expected one result, got {len(resp)}")
-
-        read_response = BluetoothGATTRead.from_pb(resp[0])
+        read_response = BluetoothGATTRead.from_pb(resp)
 
         return bytearray(read_response.data)
 
@@ -612,16 +656,20 @@ class APIClient:
         address: int,
         handle: int,
         data: bytes,
+        timeout: float = DEFAULT_BLE_TIMEOUT,
     ) -> None:
-        self._check_authenticated()
-
         req = BluetoothGATTWriteDescriptorRequest()
         req.address = address
         req.handle = handle
         req.data = data
 
-        assert self._connection is not None
-        await self._connection.send_message(req)
+        await self._send_bluetooth_message_await_response(
+            address,
+            handle,
+            req,
+            BluetoothGATTWriteResponse,
+            timeout=timeout,
+        )
 
     async def bluetooth_gatt_start_notify(
         self,
@@ -629,7 +677,13 @@ class APIClient:
         handle: int,
         on_bluetooth_gatt_notify: Callable[[int, bytearray], None],
     ) -> Callable[[], Coroutine[Any, Any, None]]:
-        self._check_authenticated()
+
+        await self._send_bluetooth_message_await_response(
+            address,
+            handle,
+            BluetoothGATTNotifyRequest(address=address, handle=handle, enable=True),
+            BluetoothGATTNotifyResponse,
+        )
 
         def on_msg(msg: message.Message) -> None:
             if isinstance(msg, BluetoothGATTNotifyDataResponse):
@@ -638,16 +692,13 @@ class APIClient:
                     on_bluetooth_gatt_notify(handle, bytearray(notify.data))
 
         assert self._connection is not None
-        await self._connection.send_message_callback_response(
-            BluetoothGATTNotifyRequest(address=address, handle=handle, enable=True),
-            on_msg,
-        )
+        remove_callback = self._connection.add_message_callback(on_msg)
 
         async def stop_notify() -> None:
             if self._connection is None:
                 return
 
-            self._connection.remove_message_callback(on_msg)
+            remove_callback()
 
             self._check_authenticated()
 
